##==============================================================================
##	File:		BIOS150.s
##	Copyright:	Copyright 2010 UC Berkeley
##==============================================================================

##==============================================================================
##	Section:	License
##==============================================================================
##	Copyright (c) 2009-2010, Regents of the University of California
##	All rights reserved.
##
##	Redistribution and use in source and binary forms, with or without modification,
##	are permitted provided that the following conditions are met:
##
##		- Redistributions of source code must retain the above copyright notice,
##			this list of conditions and the following disclaimer.
##		- Redistributions in binary form must reproduce the above copyright
##			notice, this list of conditions and the following disclaimer
##			in the documentation and/or other materials provided with the
##			distribution.
##		- Neither the name of the University of California, Berkeley nor the
##			names of its contributors may be used to endorse or promote
##			products derived from this software without specific prior
##			written permission.
##
##	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
##	ANY EXPRESS OR InIMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE InIMPLIED
##	WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
##	DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
##	ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
##	(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
##	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
##	ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
##	(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
##	SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
##==============================================================================

##==============================================================================
##	Description
##==============================================================================
## Console monitor for "poking", "peaking", file transfer, and execution.
##
## At the prompt(">"), enter an address in hex (0-9, a-f), to move to that 
## memory address.  The monitor will display the value in hex of that memory 
## address.  Once you are at a memory address, you can:
##	a.) Change its value by entering in a new in hex (8 characters required).  
##		Once you type in a new value, enter return to deposit the value and move
##		on to open the next memory address.  Alternatively, to enter a new 
##		value, you can just enter return and monitor will step to the next word 
##		memory address.
##	b.) Enter 'X' to begin program execution at that address.
##	c.) Enter 'fget'.  This will initiate a TFTP read with your PC.  All data 
##	read will be packed into sequential 32b words in memory, starting at the 
##	address that you are currently at.  The file that the TFTP client looks for 
##	when performing the read is 'fpgadat' (i.e. the name is fixed).
##	
## At anytime if you would like to move to a particular memory address, you can 
## do so by entering '<', this will get you a new prompt so that you can start 
## over.
##==============================================================================

##==============================================================================
##	Revision History
##==============================================================================
## Revision D - The Austins & Chris Fletcher (5/5/10)
##  Bug fixes:
##    1.) The last 32b word in fget commands will no longer get dropped.
##    2.) Fixed bit masking on register v0 when identifying keystrokes in 
##        'outval'
##    3.) Added masking to SpinInLoop and SpinOutLoop in the TFTP client code
## Revision C - Chris Fletcher & John Wawrzynek (4/2/10)
##  Added missing lw delay slots
## Revision B - Chris Fletcher & John Wawrzynek (4/2/10)
##	Added boot monitor code.  Boot Monitor based on Spring 2009 implementation 
##	by John Wawrzynek.
## Revision A - Chris Fletcher (3/30/10)
##	Initial TFTP Client implementation.  Supports RRQ but not WRQ.
##==============================================================================

################################################################################
#	Header
################################################################################
		.text
################################################################################

################################################################################
#	Main
################################################################################
main:
	j		__boot
	nop
################################################################################

################################################################################

###################################################################################################
###################################################################################################
##	Section: Command-line Boot Monitor
###################################################################################################
###################################################################################################

################################################################################
#	Helpers
################################################################################
#
# console I/O routines
#
inchar: 						# include automatic echo
	addiu   $sp, $sp, -4    	# save return address
	sw      $ra, 0($sp)
	ori     $t0, $0, 0xffff 	# form control register address
	sll     $t0, $t0, 16    	# ffff0000
L0:     
	lw      $t1, 0($t0)     	# get control register
	nop
	andi    $t1, $t1, 0x1   	# check if ready
	beq     $t1, $0, L0     	# then wait if not
	nop
	lw      $v0, 4($t0)     	# get the byte
	nop
	addu    $a0, $0, $v0    	# copy byte for echo
	jal     outchar         	# send back the byte
	nop
	lw      $ra, 0($sp)     	# restore return address
	addiu	$sp, $sp, 4     	#  and stack pointer
	jr      $ra             	# return
	nop

#
# Accepts string of 8 ASCII characters in the range of 0--9 and A--F,
#  first char is passed in.
#
inword:
	addiu   $sp, $sp, -12   	# save return address, s0, s1
	sw      $ra, 0($sp)
	sw      $s0, 4($sp)
	sw      $s1, 8($sp)
	ori		$s1, $0, 0x0		# s1 holds the result
	ori     $s0, $0, 8      	# char loop counter
inword0:
	addiu   $v0, $v0, -48   	# subtract ascii 0 (0x30)
	ori     $t1, $0, 9      	# compare to 9
	slt     $t0, $t1, $v0   	# if x>9
	beq     $t0, $0, inword1 	# branch if in 0--9
	nop
	addiu   $v0, $v0, -7    	# adjust for A--F
inword1:
	addiu	$t0, $s0, -1    	# form shift amount
inword3:
	andi    $v0, $v0, 0x0f
	addu    $s1, $s1, $v0   	# put new value in result
	addiu   $s0, $s0, -1    	# dec char loop counter
	beq     $s0, $0, inword4
	nop
	sll		$s1, $s1, 4
	jal     inchar          	# get next ascii char
	nop
	beq     $0, $0, inword0
	nop
inword4:
	ori		$v0, $s1, 0x0		# return result
	lw      $s1, 8($sp)
	lw      $s0, 4($sp)
	lw      $ra, 0($sp)    		# restore return address
	addiu   $sp, $sp, 12    	#  and stack pointer
	jr      $ra             	# return
	nop

outchar:
	addiu   $sp, $sp, -4    	# save return address
	sw      $ra, 0($sp)
	ori     $t0, $0, 0xffff 	# form control register address
	sll     $t0, $t0, 16    	# ffff0000
	L1:     
	lw      $t1, 8($t0)     	# get control register
	nop
	andi    $t1, $t1, 0x1   	# check if ready
	beq     $t1, $0, L1     	# then wait if not
	nop
	sw      $a0, 12($t0)    	# put the byte
	lw      $ra, 0($sp)     	# restore return address
	addiu   $sp, $sp, 4     	#  and stack pointer
	jr      $ra             	# return
	nop

#
# Outputs a 32-bit word as ASCII string
#
outword:
	addiu   $sp, $sp, -12   	# save return address, s0, s1
	sw      $ra, 0($sp)
	sw      $s0, 4($sp)
	sw      $s1, 8($sp)
	ori		$s1, $a0, 0x0		# s1 holds output word
	ori     $s0, $0, 8      	# s0 holds char counter
outword0:
	ori		$t1, $s1, 0x0		# copy output word
	addiu   $t0, $s0, -1    	# form shift amount
outword1:
	beq     $t0, $0, outword2 	# exit for shift loop
	nop
	sra     $t1, $t1, 4     	# shift nibble at a time
	addiu   $t0, $t0, -1    	# dec shift counter
	beq     $0, $0, outword1 	# keep shifting
	nop
outword2:
	andi    $t1, $t1, 0x000f 	# isolate low nibble
	addiu   $t1, $t1, 48    	# add ascii '0'
	ori     $t2, $0, 57     	# compare to ascii '9'
	slt     $t0, $t2, $t1   	# if x > 57 don't
	beq     $t0, $0, outword3 	# skip
	nop
	addiu   $t1, $t1, 7     	# adjust for A--F
outword3:
	ori		$a0, $t1, 0x0		# send char to console
	jal     outchar
	nop
	addiu   $s0, $s0, -1    	# dec char loop counter
	bne     $s0, $0, outword0 	# keep going
	nop
	lw      $s1, 8($sp)
	lw      $s0, 4($sp)
	lw      $ra, 0($sp)     	# restore return address
	addiu   $sp, $sp, 12    	#  and stack pointer
	jr      $ra             	# return
	nop
################################################################################

################################################################################
#	Boot Monitor Entry Point
#	Implements command-line boot monitor behavior
################################################################################

# Output Version Number
__boot:			
	li		$sp, 0x7ffffffc		# initialize stack pointer
	ori     $a0, $0, 0x56   	# 'V'
	jal     outchar
	nop
	ori     $a0, $0, 4      	# Version Number
	jal     outword
	nop

	# Output Prompt
prompt: 				
	ori     $a0, $0, 0xa    	# '\n'
	jal     outchar
	nop
	ori      $a0, $0, 0x3e   	# '>'
	jal     outchar
	nop

	# Input Address
	jal     inchar          	# get first char
	nop
	ori     $a0, $v0, 0x0
	jal     inword
	nop
	ori     $s0, $v0, 0x0		# save address


	# Output Value surrounded by a space
outval: 
	ori     $a0, $0, 0x20   	# ' '
	jal     outchar
	nop
	lw      $a0, 0($s0)     	# value<-mem[address]
	jal     outword         	# send to console
	nop
	ori     $a0, $0, 0x20   	# ' '
	jal     outchar
	nop

	# Now expecting '\n', 'X', '<', or new value
	jal     inchar
	nop
	
	addiu	$sp, $sp, -4
	sw		$v0, 0($sp)
	
	ori     $t0, $0, 0x58   	# 'X'
	beq     $t0, $v0, exec
	nop
	
	lw		$v0, 0($sp)
	nop
	ori     $t0, $0, 0x3c   	# '<'
	beq     $t0, $v0, prompt
	nop
	
	lw		$v0, 0($sp)
	nop	
	ori     $t0, $0, 0xa    	# '\n'
	beq     $t0, $v0, peak
	nop
	
	lw		$v0, 0($sp)
	nop	
	ori     $t0, $0, 0x66   	# 'f'
	beq     $t0, $v0, ethernet
	nop
	
	addiu	$sp, $sp, 4
	
	# Assume new value
poke:   						# Set memory
	ori		$a0, $t0, 0x0		# pass in first char
	jal     inword          	# get the value
	nop
	sw      $v0, 0($s0)     	# mem[address]<-value
	jal     inchar          	# eat the '\n'
								# fall through
	nop
					   
peak:   						# show address
	addiu   $s0, $s0, 4     	# incr memory address
	ori     $a0, $0, 0x3e   	# '>'
	jal     outchar
	nop
	ori		$a0, $s0, 0x0		# output new adress
	jal     outword
	nop
	ori     $a0, $0, 0x20   	# and a space
	jal     outchar
	nop
	beq     $0, $0, outval  	# show value and wait
	nop

exec:   						# Start executing
	jal     exec0           	# form return address
	nop
	exec0:  
	addiu   $ra, $ra, 8
	jr      $s0             	# jump to it
	nop
	beq     $0, $0, prompt  	# return to here
	nop
	
ethernet:
	jal     inchar
	nop
	beq     $v0, 0x67, __fget	# 'g'
	nop
	beq     $v0, 0x70, __fput	# 'p'
	nop
	j		poke				# it wasn't ethernet at all -- just a hex character!
	nop
	
__fget:
	jal     inchar
	nop
	bne     $v0, 0x65, prompt	# 'e'
	nop
	jal     inchar
	nop
	bne     $v0, 0x74, prompt	# 't'
	nop
	
	# start the TFTP RRQ
	ori		$a0, $0, 0			# opcode for fget/RRQ
	ori 	$a1, $s0, 0x0		# store the address
	jal		__tftp
	nop
	j		prompt
	nop
	
__fput:
	## TODO: Implement
	j		prompt
	nop
################################################################################

###################################################################################################
###################################################################################################
##	Section: TFTP Client
###################################################################################################
###################################################################################################

################################################################################
#	Global Assignments (Notes & Comments)
################################################################################
# 	Reserved Addresses
#	The following data memory addresses are used by the TFTPClient
#	0x10010000 	TX bundle buffer
#	0x10010100 	tftpFSM args
#	0x10010200	RX bundle buffer
#	0x10010300	Reserved:
#		0x10010300 - destination address for a RRQ
#		0x10010304 - # bytes to send in a WRQ
#	0x10010400	Packet (IP/UDP header) buffer
#	0x10010590	Payload (TFTP) buffer
 
# 	Register Usage
#	s0	opcode		+0
#	s1  src TID		+4
#	s2	dest TID	+8
#	s3	seqNum		+12
#	s4	data length	+16

#	tftpFSM state encodings:
#	Init:			0
#	Connect:		1
#	RRQLoop:		2
#	WRQLoop:		3
#	Terminate:		4
	
#	Constants
#	MAX_LEN: 		524

#	TFTP Constants
#	Source TID		0x411
################################################################################

################################################################################
#	Helpers
################################################################################
htons:
	#	Args:
	#	a0	input
	addiu	$v1, $a0, 0
	andi	$v1, $v1, 0x00FF	# LSB
	sll		$v1, $v1, 8
	srl		$v0, $a0, 8
	or		$v0, $v0, $v1
	jr		$ra	
	nop

ntohs:
	#	Args:
	#	a0	input
	addiu	$v1, $a0, 0
	andi	$v1, $v1, 0x00FF	# LSB
	sll		$v1, $v1, 8
	srl		$v0, $a0, 8
	or		$v0, $v0, $v1
	jr		$ra	
	nop

shiftWrite:
	#	Args:
	#	a0	input
	srl		$a0, $a0, 8
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)	
	jal		write8b
	nop
	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4
	jr		$ra
	nop

write8b:
	#	Args:
	#	a0	input
spinOut:
	addiu	$t0, $0, 0xffff		# form ControlOutReg address
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x0018
spinOutLoop:
	lw		$t1, 0($t0)			# Load the ControlOutReg status word
	nop
	andi	$t1, $t1, 1
	beq		$t1, $0, spinOutLoop
	nop
	
	addiu	$t0, $0, 0xffff		# form DataOutReg address
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x001c
	andi	$a1, $a0, 0x00ff
	sw		$a1, 0($t0)	
	jr		$ra
	nop

write16b:
	#	Args:
	#	a0	input
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)	
	jal		write8b
	nop
	jal		shiftWrite
	nop
	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4	
	jr		$ra
	nop	

write32b:
	#	Args:
	#	a0	input
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)	
	jal		write8b
	nop
	jal		shiftWrite
	nop
	jal		shiftWrite
	nop
	jal		shiftWrite
	nop
	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4	
	jr		$ra
	nop

#
# Forms a 16b word starting from 2 bytes stored at sequential addresses.
# Follows the format:
#	MSB {addr +0, addr +4} LSB
#
form16b:
	#	Args:
	#	a0	start address
	lw		$t0, 4($a0)
	nop
	sll		$t1, $t0, 8
	lw		$t0, 0($a0)
	nop
	or		$a0, $t1, $t0
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)
	jal		ntohs
	nop
	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4	
	jr		$ra
	nop
	
#
# Forms a 32b word from 4 bytes "".  Follows the format:
#	MSB {addr+12, addr+8, addr+4, addr+0} LSB
#
form32b:
	#	Args:
	#	a0	start address
	lw		$v0, 0($a0)
	nop
	
	lw		$t1, 4($a0)
	nop	
	sll		$t1, $t1, 8
	or		$v0, $v0, $t1

	lw		$t1, 8($a0)
	nop
	sll		$t1, $t1, 16
	or		$v0, $v0, $t1

	lw		$t1, 12($a0)
	nop	
	sll		$t1, $t1, 24
	or		$v0, $v0, $t1
	jr		$ra
	nop

#
#	Args:
#	a0		end address
#
commitPacket:
	addiu	$a2, $a0, 0			# load end address
	li		$a0, 0x10010400		# load start address
	addiu	$a0, $a0, 128
	li		$a1, 0x10010300		# load write address
	lw		$a1, 0($a1)
	nop
#
# Packages the bytes received from a packet into 32b words and writes them to 
# the indicated address range.  This function cuts off when there is < 32b left 
# to write, thus in order to write every bit of data, the address range must be 
# % 32 = 0.
#	Args:
#	a0	start address: where is the start of valid data?	
#	a1	write address: where should data be written to?
#	a2	end address: what is the last address that valid data is in?
#	
commitPacketBody:
	addiu	$t0, $a0, 12		
	slt		$t0, $t0, $a2		# if the next 32b are within the remaining range
	beq		$t0, $0, commit_end
	nop

	# build the 32b word from sequential 8b words
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)
	jal		form32b
	nop
	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4	
	
	sw		$v0, 0($a1)			# commit word
	
	# increment pointers
	addiu	$a0, $a0, 16		# 4 consequtive addresses
	addiu	$a1, $a1, 4
	
	j		commitPacketBody
	nop
	
commit_end:
	li		$t0, 0x10010300		# load write address
	sw		$a1, 0($t0)
	jr		$ra
	nop
################################################################################

################################################################################
#	TFTPClient Entry Point
################################################################################
__tftp:
	#	Args:
	#	a0		Command (0 - fget, 1 - fput)	+0
	#	a1		Address (where to write data on an RRQ or start to read data 
	#			from in a WRQ)
	#	a2		Size.  The amount of data to write.  Only used with WRQ.
	li		$t0, 0x10010300
	sw		$a1, 0($t0)
	sw		$a2, 4($t0)
	
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)
	jal		tftpFSM
	nop
	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4	
	jr		$ra
	nop
################################################################################

################################################################################
#	tftpFSM
#	Implements the TFTP protocol
################################################################################
tftpFSM:
	#	Args:
	#	a0		Command (0 - fget, 1 - fput)	+0
	#
	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x100
	sw		$a0, 0($t0)

	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)	
	
	addiu	$t8, $0, 0		# Set the Init state
		
	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x000
	addiu	$t4, $0, 0x411	# TX bundle src TID	
	sw		$t4, 4($t0)
	addiu	$t4, $0, 1		# TX bundle sequence number	
	sw		$t4, 12($t0)

tftpMainLoop:
	beq		$t8, 0, STATE_Init
	nop
	beq		$t8, 1, STATE_Connect
	nop
	beq		$t8, 2, STATE_RRQLoop
	nop
	beq		$t8, 3, STATE_WRQLoop
	nop
	beq		$t8, 4, STATE_Terminate
	nop

STATE_Init:
	beq		$a0, 0, Init_GET
	nop
	
	Init_GET:
		addiu	$t0, $0, 0x1001
		sll		$t0, $t0, 16
		ori		$t0, $t0, 0x000
		addiu	$s0, $0, 1
		lw		$s1, 4($t0)
		nop		
		# We don't need to set the dst TID!
		lw		$s3, 12($t0)		
		nop
		
		jal		TunnelIn
		nop

		addiu	$t0, $0, 0x1001
		sll		$t0, $t0, 16
		ori		$t0, $t0, 0x000
		sw		$s1, 4($t0)		
		sw		$s2, 8($t0)
		sw		$s3, 12($t0)
		
		# next state
		addiu	$t8, $0, 1

		j		tftpMainLoop
		nop

	Init_PUT:	
		# TODO: Implement
		j		tftpMainLoop
		nop

STATE_Connect:
	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x200
	lw		$s1, 4($t0)
	nop	
	lw		$s2, 8($t0)
	nop	
	lw		$s3, 12($t0)
	nop
	
	jal		TunnelOut
	nop
	
	# Check the dst TID to see if it matches the FPGA's
	addiu	$t1, $0, 0x1001
	sll		$t1, $t1, 16
	ori		$t1, $t1, 0x000	
	lw		$t3, 4($t1)			# FPGA's TID
	nop	
	bne		$t3, $s2, STATE_Connect # retry if not	
	nop

	# Sequence number logic
	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x200	
	bne		$s3, 1, STATE_Connect # If sequence number isn't 1, wait...
	nop
	
	# commit receive stats
	sw		$s1, 4($t0)
	sw		$s2, 8($t0)
	sw		$s3, 12($t0)	
	
	# transfer src->dst TIDs	
	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x000
	sw		$s1, 8($t0)

	# commit packet if RRQ
	bne		$v0, 1, Connect_NoConnect
	nop
	addiu	$a0, $v1, 0
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)
	jal		commitPacket
	nop
	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4
	
	Connect_NoConnect:
		# termination case
		beq		$v0, 0, Connect_NS	# if packet is ACK, we don't need to check packet length
		nop
		bne		$s4, 524, LastPacket
		nop
	
	Connect_NS:						# next state
		addiu	$t0, $0, 0x1001
		sll		$t0, $t0, 16
		ori		$t0, $t0, 0x100		
		lw		$t0, 0($t0)
		nop	
		beq		$t0, 0, Connect_NS_RRQLoop
		nop	
		addiu	$t8, $0, 3			# else case (WRQ for now)
		j		tftpMainLoop
		nop	

	Connect_NS_RRQLoop:
		addiu	$t8, $0, 2
		j		tftpMainLoop
		nop

STATE_RRQLoop:

	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x000
	addiu	$s0, $0, 4			# set ACK
	lw		$s1, 4($t0)
	nop	
	lw		$s2, 8($t0)
	nop	
	lw		$s3, 12($t0)
	nop
	
	jal		TunnelIn
	nop

	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x000
	sw		$s1, 4($t0)		
	sw		$s2, 8($t0)
	sw		$s3, 12($t0)		

RRQLoop_Listen:
	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x200
	lw		$s1, 4($t0)
	nop	
	lw		$s2, 8($t0)
	nop	
	lw		$s3, 12($t0)
	nop
	
	jal		TunnelOut
	nop

	# retry if packet isn't data...
	bne		$v0, 1, RRQLoop_Listen
	nop	
	
	# Check the dst TID & src TID
	addiu	$t1, $0, 0x1001
	sll		$t1, $t1, 16
	ori		$t1, $t1, 0x000	
	lw		$t3, 4($t1)			# FPGA's TID
	nop	
	lw		$t4, 8($t1)			# PC's TID
	nop	
	bne		$t3, $s2, RRQLoop_Listen # retry if dst mismatches	
	nop
	bne		$t4, $s1, RRQLoop_Listen # retry if src mismatches
	nop
	
	# Sequence number logic
	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x200	
	lw		$t3, 12($t1)		# FPGA's current sequence number
	nop	
	addiu	$t3, $t3, 1			# Increment sequence number		
	bne		$t3, $s3, RRQLoop_Listen # If the sequence number doesn't match expected, wait...
	nop
	# Store new seq numbers
	sw		$t3, 12($t1)		
	sw		$s3, 12($t0)
	
	# commit receive stats
	sw		$s1, 4($t0)		
	sw		$s2, 8($t0)
	sw		$s3, 12($t0)	
	sw		$s4, 16($t0)

	# commit packet
	addiu	$a0, $v1, 0
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)
	jal		commitPacket
	nop
	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4	
	
	bne		$s4, 524, LastPacket
	nop
	# by default, don't change states
	j		tftpMainLoop
	nop

LastPacket:
	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x000	
	addiu	$s0, $0, 4			# set ACK
	lw		$s1, 4($t0)
	nop	
	lw		$s2, 8($t0)
	nop	
	lw		$s3, 12($t0)
	nop
	
	jal		TunnelIn
	nop

	addiu	$t0, $0, 0x1001
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x000
	sw		$s1, 4($t0)		
	sw		$s2, 8($t0)
	sw		$s3, 12($t0)	

	# next state
	addiu	$t8 $0, 4			# terminate
	
	j		tftpMainLoop
	nop	

STATE_WRQLoop:
	# TODO: Implement
STATE_Terminate:
	# TODO: Implement

	lw		$ra, 0($sp)
	nop	
	addiu	$sp, $sp, 4	
	
	jr		$ra
	nop
################################################################################

################################################################################
#	TunnelIn
#	Writes a packet from the CPU->EMAC
#	Returns: -
################################################################################
TunnelIn:
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)

	addiu	$s5, $0, 0x1001 	# Packet (IP/UDP header) buffer
	sll		$s5, $s5, 16
	ori		$s5, $s5, 0x400	
	addiu	$s6, $0, 0x1001 	# Payload (TFTP) buffer	
	sll		$s6, $s6, 16
	ori		$s6, $s6, 0x590
	addiu	$s7, $0, 0			# Packet length

	# build the TFTP header
	addiu	$a0, $s0, 0			# Build the TFTP opcode
	jal		htons
	nop	
	sw		$v0, 0($s6)			# Store the TFTP opcode

	beq		$s0, 1, RWRREQ		# Is packet read request?
	nop		
	beq		$s0, 2, RWRREQ		# Is packet write request?
	nop
	j		DATAACK
	nop

RWRREQ:
	addiu	$s7, $s7, 16		# length += opcode + file + mode
	j		IP
	nop

DATAACK:
	addiu	$s7, $s7, 4			# length += opcode + block# 
	addiu	$a0, $s3, 0			# Build the TFTP block number
	jal		htons
	nop	
	sw		$v0, 4($s6)			# Store the TFTP block number
	
IP:	
	# build the IP header
	addiu	$t4, $0, 4			# v
	sll		$t4, $t4, 4
	ori		$t4, $t4, 5			# hl	
	sw		$t4, 0($s5)			# store {hl, v}

	addiu	$t4, $0, 0			# tos
	sw		$t4, 4($s5)			# store tos

	addiu	$a0, $s7, 0			# len += TFTP header length
	addiu	$a0, $a0, 20		# len += IP header length
	addiu	$a0, $a0, 8			# len += UDP header length
	jal		htons
	nop
	sw		$v0, 8($s5)			# store len

	addiu	$t4, $0, 0x2501		# identification
	sw		$t4, 12($s5)		# store identification

	addiu	$t4, $0, 0			# offset
	sw		$t4, 16($s5)		# store offset	

	addiu	$t4, $0, 128		# ttl
	sw		$t4, 20($s5)		# store ttl

	addiu	$t4, $0, 0x11		# prot (for UDP)
	sw		$t4, 24($s5)		# store prot

	addiu	$t4, $0, 0			# checksum (initially)
	sw		$t4, 28($s5)		# store checksum

	addiu	$t4, $0, 0xcb4d 	# FPGA IP "169.254.77.203"
	sll		$t4, $t4, 16
	ori		$t4, $t4, 0xfea9
	sw		$t4, 32($s5)		# store src IP
	
	addiu	$t4, $0, 0xcc4d 	# PC IP "169.254.77.204"
	sll		$t4, $t4, 16
	ori		$t4, $t4, 0xfea9
	sw		$t4, 36($s5)		# store dst IP

#
#	Calculate the checksum (optimized for sending only TFTP packets)
#
Checksum_Fixed:
	beq		$s0, 1, CS_RRQ		
	nop		
	beq		$s0, 4, CS_ACK	
	nop
CS_RRQ:
	addiu	$t4, $0, 0x084a		# RRQ checksum
	j		CS_Commit
	nop
CS_ACK:
	addiu	$t4, $0, 0x144a		# ACK checksum
	j		CS_Commit
	nop
CS_Commit:
	sw		$t4, 28($s5)		# store checksum
	
#
# sendto call
#
Write_IP:
	# write the IP header
	lw		$a0, 0($s5)
	nop	
	jal		write8b
	nop	
	lw		$a0, 4($s5)
	nop	
	jal		write8b
	nop
	lw		$a0, 8($s5)
	nop	
	jal		write16b
	nop
	lw		$a0, 12($s5)
	nop	
	jal		write16b
	nop
	lw		$a0, 16($s5)
	nop
	jal		write16b
	nop
	lw		$a0, 20($s5)
	nop
	jal		write8b
	nop
	lw		$a0, 24($s5)
	nop
	jal		write8b
	nop
	lw		$a0, 28($s5)
	nop	
	jal		write16b
	nop
	lw		$a0, 32($s5)
	nop	
	jal		write32b
	nop
	lw		$a0, 36($s5)
	nop	
	jal		write32b
	nop

Write_UDP:
	# write the UDP header
	addiu	$a0, $s1, 0
	jal		htons
	nop
	addiu	$a0, $v0, 0
	jal		write16b			# src TID
	nop

	beq		$s0, 1, UDP_RRQ	
	nop		
	beq		$s0, 2, UDP_RRQ	
	nop
	addiu	$a0, $s2, 0			# We are processing Data/ACK packets
	jal		htons
	nop
	j		UDP_DST
	nop

UDP_RRQ:
	addiu	$a0, $0, 69			# UDP port
	jal		htons
	nop

UDP_DST:
	addiu	$a0, $v0, 0			# From either UDP_RRQ or Data/Ack packet dst
	jal		write16b			# dst TID
	nop

	addiu	$a0, $0, 8			# len += UDP length
	addu	$a0, $a0, $s7		# len += TFTP length
	jal		htons
	nop
	addiu	$a0, $v0, 0
	jal		write16b			# len
	nop

	addiu	$a0, $0, 0
	jal		write16b			# checksum
	nop

Write_TFTP:
	lw		$a0, 0($s6)
	nop	
	jal		write16b			# opcode
	nop	

	# copy the payload to the packet buffer
	beq		$s0, 1, TFTP_RRQ	
	nop		
	beq		$s0, 2, TFTP_RRQ	
	nop	
	beq		$s0, 3, TFTP_DATA	
	nop
	beq		$s0, 4, TFTP_ACK	
	nop

TFTP_RRQ:
	# read/write requests
	
	# remote file
	addiu	$a0, $0, 'f'		
	jal		write8b		
	nop
	addiu	$a0, $0, 'p'		
	jal		write8b				
	nop
	addiu	$a0, $0, 'g'		
	jal		write8b			
	nop
	addiu	$a0, $0, 'a'		
	jal		write8b			
	nop
	addiu	$a0, $0, 'd'		
	jal		write8b			
	nop
	addiu	$a0, $0, 'a'		
	jal		write8b			
	nop
	addiu	$a0, $0, 't'		
	jal		write8b			
	nop

	# terminator
	addiu	$a0, $0, 0	
	jal		write8b			
	nop	

	# remote mode
	addiu	$a0, $0, 'o'		
	jal		write8b			
	nop
	addiu	$a0, $0, 'c'		
	jal		write8b			
	nop
	addiu	$a0, $0, 't'		
	jal		write8b			
	nop
	addiu	$a0, $0, 'e'		
	jal		write8b			
	nop
	addiu	$a0, $0, 't'		
	jal		write8b			
	nop
	
	j		TunnelInDone
	nop

TFTP_DATA:
	# data packets
	# TODO: implement
	j		TunnelInDone
	nop

TFTP_ACK:
	# ack
	addiu	$a0, $s3, 0
	jal		htons
	nop
	addiu	$a0, $v0, 0
	jal		write16b	
	nop
	
TunnelInDone:
	addiu	$a0, $0, 0x100		# write the terminator byte
	addiu	$t0, $0, 0xffff		# form DataOutReg address
	sll		$t0, $t0, 16
	ori		$t0, $t0, 0x001c
	sw		$a0, 0($t0)	

	lw		$ra, 0($sp)
	nop
	addiu	$sp, $sp, 4		
	
	jr		$ra
	nop
################################################################################
	
################################################################################
#	TunnelOut
#	Polls the EMAC->CPU path and receives/processes a single packet
#	Returns:
#		v0:
#			1 - the received packet was a valid TFTP data packet
#			0 - "" ACK packet
#		v1: the address of the last valid/written byte of data
################################################################################
TunnelOut:
	addiu	$sp, $sp, -4
	sw		$ra, 0($sp)

#
# recvfrom call
#
spinIn:
	addiu	$s5, $0, 0x1001 	# Packet (IP/UDP header) buffer
	sll		$s5, $s5, 16
	ori		$s5, $s5, 0x400		
	
	addiu	$t3, $0, 0			# clear the start bit

	addiu	$t0, $0, 0xffff		# form ControlInReg
	sll		$t0, $t0, 16	
	ori		$t0, $t0, 0x0010
	
spinInLoop:
	lw		$t1, 0($t0)			# load Control bit
	nop
	andi	$t1, $t1, 1			# mask out the control bit
	beq		$t1, 0, spinInLoop
	nop
	lw		$t1, 4($t0)			# load data byte
	nop
	srl		$t2, $t1, 8
	andi	$t2, $t2, 0x1		# isolate the LSB
	bne		$t2, 0, recvTag		# encountered a start/end frame word
	nop
recvData:
	sw		$t1, 0($s5)			# encountered a normal data word
	addiu	$s5, $s5, 4
	j		spinInLoop
	nop
recvTag:	
	andi	$t1, $t1, 0xFF
	beq		$t1, 1, recvStart
	nop
	beq		$t1, 2, recvStopGood
	nop	
	beq		$t1, 4, recvStopBad
	nop
	
recvStart:
	# We have to
	#	a.) Increment the packet buffer pointer
	#	b.) Check and handle the packet overflow case
	beq		$t3, 1, overflow	# if we receive two starts in a row...
	nop
	addiu	$t3, $0, 1			# set the start bit
	j		spinInLoop
	nop		
overflow:
	addiu	$s5, $0, 0x1001 	# throw out current packet and restart
	sll		$s5, $s5, 16
	ori		$s5, $s5, 0x400		
	j		spinInLoop
	nop
	
recvStopGood:
	addiu	$t3, $0, 0			# clear the start bit
	j		recvDone
	nop	
recvStopBad:
	addiu	$t3, $0, 0			# clear the start bit
	j		spinInLoop
	nop

recvDone:
	# set the return register to the last byte address of valid data
	addiu	$sp, $sp, -4
	sw		$s5, 0($sp)
	
	addiu	$s5, $0, 0x1001 	# Reset buffer pointer
	sll		$s5, $s5, 16
	ori		$s5, $s5, 0x400
	
	# ip/udp filtering
	lw		$t0, 36($s5)		# look at the ip protocol field
	nop
	bne		$t0, 0x11, spinIn	# if it isn't a UDP packet... wait for another packet
	nop

	addiu	$a0, $s5, 80		# save the src TID field
	jal		form16b
	nop
	addiu	$s1, $v0, 0

	addiu	$a0, $s5, 88		# save the dst TID field
	jal		form16b
	nop
	addiu	$s2, $v0, 0
	
	addiu	$a0, $s5, 96		# save the data len
	jal		form16b
	nop
	addiu	$s4, $v0, 0	

	lw		$s0, 116($s5)		# save the TFTP opcode field
	nop

	addiu	$a0, $s5, 120		# block #/sequence number
	jal		form16b
	nop
	addiu	$s3, $v0, 0

	# return value				Was the block data?
	beq		$s0, 3, TunnelOutDone1
	nop
	
TunnelOutDone0:
	addiu	$v0, $0, 0

	lw		$ra, 4($sp)		
	nop
	addiu	$sp, $sp, 8	
	
	jr		$ra
	nop
TunnelOutDone1:
	addiu	$v0, $0, 1
	
	lw		$v1, 0($sp)			# load the end address
	nop
	lw		$ra, 4($sp)
	nop
	addiu	$sp, $sp, 8	
	
	jr		$ra
	nop
################################################################################
